/*	$NetBSD$	*/
/*	$NetBSD/i386: isadma.c,v 1.42 1999/03/22 07:06:09 mycroft Exp $	*/

/*
 * [NetBSD for NEC PC-98 series]
 *  Copyright (c) 1994, 1995, 1996, 1997, 1998, 1999
 *	NetBSD/pc98 porting staff. All rights reserved.
 */

/*-
 * Copyright (c) 1997, 1998 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Jason R. Thorpe of the Numerical Aerospace Simulation Facility,
 * NASA Ames Research Center.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the NetBSD
 *	Foundation, Inc. and its contributors.
 * 4. Neither the name of The NetBSD Foundation nor the names of its
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Device driver for the CBus on-board DMA controller.
 */

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/proc.h>
#include <sys/device.h>
#include <sys/malloc.h>

#include <vm/vm.h>

#include <machine/bus.h>

#include <pc98/cbus/cbusreg.h>
#include <pc98/cbus/cbusvar.h>
#include <pc98/cbus/cbusdmavar.h>
#include <pc98/cbus/cbusdmareg.h>

/*
 * Entry points for CBus DMA.  These are mostly wrappers around
 * the generic functions that understand how to deal with bounce
 * buffers, if necessary.
 */
struct i386_bus_dma_tag isa_bus_dma_tag = {
	CBUS_DMA_BOUNCE_THRESHOLD,
	_cbus_bus_dmamap_create,
	_cbus_bus_dmamap_destroy,
	_cbus_bus_dmamap_load,
	_cbus_bus_dmamap_load_mbuf,
	_cbus_bus_dmamap_load_uio,
	_cbus_bus_dmamap_load_raw,
	_cbus_bus_dmamap_unload,
	_cbus_bus_dmamap_sync,
	_cbus_bus_dmamem_alloc,
	_bus_dmamem_free,
	_bus_dmamem_map,
	_bus_dmamem_unmap,
	_bus_dmamem_mmap,
};


struct cbus_mem *cbus_mem_head;

/*
 * High byte of DMA address is stored in this DMAPG register for
 * the Nth DMA channel.
 */
#define	DMA_NCHAN	4

/* dma auto incremental bank address control register */
#define	DMAABA_64K	0x00
#define	DMAABA_1M	0x04
#define	DMAABA_16M	0x0c

/* dma bank address register */
static bus_addr_t dmapageportiat[4] = {0x26, 0x20, 0x22, 0x24};

/* 8237AC registers */
static int dmapageport[2][4] = {
	{0x0, 0x1, 0x2, 0x3}, {0x0, 0x1, 0x2, 0x3}
};

static u_int8_t dmamode[] = {
	/* write to device/read from device */
	DMA37MD_READ | DMA37MD_SINGLE,
	DMA37MD_WRITE | DMA37MD_SINGLE,

	/* write to device/read from device */
	DMA37MD_READ | DMA37MD_DEMAND,
	DMA37MD_WRITE | DMA37MD_DEMAND,

	/* write to device/read from device - DMAMODE_LOOP */
	DMA37MD_READ | DMA37MD_SINGLE | DMA37MD_LOOP,
	DMA37MD_WRITE | DMA37MD_SINGLE | DMA37MD_LOOP,

	/* write to device/read from device - DMAMODE_LOOPDEMAND */
	DMA37MD_READ | DMA37MD_DEMAND | DMA37MD_LOOP,
	DMA37MD_WRITE | DMA37MD_DEMAND | DMA37MD_LOOP,
};

static int _Cbus_dma_map __P((bus_space_tag_t,
		bus_space_handle_t *, bus_space_handle_t *,
		bus_space_handle_t *,
		bus_space_handle_t *, bus_space_handle_t *));

static int
_Cbus_dma_map(iot, dma1hp, dma2hp, dmapghp, dmaachp, dmaabahp)
	bus_space_tag_t iot;
	bus_space_handle_t *dma1hp, *dma2hp, *dmapghp, *dmaachp, *dmaabahp;
{
	static bus_space_handle_t dmah, dmapgh, dmaach, dmaabah;

	if (dmah == NULL)
		if (bus_space_map(iot, IO_DMA, 0, 0, &dmah) ||
		    bus_space_map_load(iot, dmah, 16, BUS_SPACE_IAT_2, 0))
			return EINVAL;
	*dma1hp = *dma2hp = dmah;

	if (dmapgh == NULL)
		if (bus_space_map(iot, IO_DMA, 0, 0, &dmapgh) ||
	    	    bus_space_map_load(iot, dmapgh,
				       BUS_SPACE_IAT_SZ(dmapageportiat),
			       	       dmapageportiat, 0))
			return EINVAL;
	*dmapghp = dmapgh;

	if (dmaach == NULL)
		if (bus_space_map(iot, 0x439, 1, 0, &dmaach))
			return EINVAL;
	*dmaachp = dmaach;

	if (dmaabah == NULL)
		if (bus_space_map(iot, 0x29, 1, 0, &dmaabah))
			return EINVAL;
	*dmaabahp = dmaabah;

	return 0;
}

static inline void _cbus_dmaunmask __P((struct cbus_dma_state *, int));
static inline void _cbus_dmamask __P((struct cbus_dma_state *, int));

static inline void
_cbus_dmaunmask(struct cbus_dma_state *cds, int chan) {
	int ochan = chan & 3;

	CBUS_DMA_MASK_CLR(cds, chan);

	/*
	 * If DMA is frozen, don't unmask it now.  It will be
	 * unmasked when DMA is thawed again.
	 */
	if (cds->cds_frozen)
		return;

	/* set dma channel mode, and set dma channel mode */
	if ((chan & 4) == 0)
		bus_space_write_1(cds->cds_bst, cds->cds_dma1h,
		    DMA1_SMSK, ochan | DMA37SM_CLEAR);
	else
		bus_space_write_1(cds->cds_bst, cds->cds_dma2h,
		    DMA2_SMSK, ochan | DMA37SM_CLEAR);
}

static inline void
_cbus_dmamask(struct isa_dma_state *cds, int chan) {
	int ochan = chan & 3;

	CBUS_DMA_MASK_SET(cds, chan);

	/*
	 * XXX Should we avoid masking the channel if DMA is
	 * XXX frozen?  It seems like what we're doing should
	 * XXX be safe, and we do need to reset FFC...
	 */

	/* set dma channel mode, and set dma channel mode */
	if ((chan & 4) == 0) {
		bus_space_write_1(cds->cds_bst, cds->cds_dma1h,
		    DMA1_SMSK, ochan | DMA37SM_SET);
		bus_space_write_1(cds->cds_bst, cds->cds_dma1h,
		    DMA1_FFC, 0);
	} else {
		bus_space_write_1(cds->cds_bst, cds->cds_dma2h,
		    DMA2_SMSK, ochan | DMA37SM_SET);
		bus_space_write_1(cds->cds_bst, cds->cds_dma2h,
		    DMA2_FFC, 0);
	}
}

/*
 * _cbus_dmainit(): Initialize the cbus_dma_state for this chipset.
 */
void
_cbus_dmainit(struct isa_dma_state *cds, bus_space_tag_t bst,
    bus_dma_tag_t dmat, struct device *dev)
{
	int i;
	u_int8_t tmp;

	cds->cds_dev = dev;

	if (cds->cds_initialized) {
		if (cds->cds_bst != bst || cds->cds_dmat != dmat)
			panic("_cbus_dmainit: inconsistent bus tags");
	} else {
		cds->cds_bst = bst;
		cds->cds_dmat = dmat;

		/*
		 * Map the registers used by the CBus DMA controller.
		 */
		if (_Cbus_dma_map(cds->cds_bst,
				 &cds->cds_dma1h, &cds->cds_dma2h,
			         &cds->cds_dmapgh,
				 &cds->cds_dmaach, &cds->cds_dmaabah))
			panic("cbusdma: can not map dma sapce\n");
		tmp = bus_space_read_1(cds->cds_bst, cds->cds_dmaach, 0) & 0xfb;
		bus_space_write_1(cds->cds_bst, cds->cds_dmaach, 0, tmp);

		/* setup auto incremental bank address control registers */
		for (i = 0; i < DMA_NCHAN; i ++)
			bus_space_write_1(cds->cds_bst,
				cds->cds_dmaabah, 0, DMAABA_16M | i);
		/*
		 * All 8 DMA channels start out "masked".
		 */
		cds->cds_masked = 0xff;

		cds->cds_initialized = 1;

		/*
		 * DRQ 4 is used to chain the two 8237s together; make
		 * sure it's always cascaded, and that it will be unmasked
		 * when DMA is thawed.
		 */
		_cbus_dmacascade(cds, 4);
	}
}

/*
 * _cbus_dmacascade(): program 8237 DMA controller channel to accept
 * external dma control by a board.
 */
int
_cbus_dmacascade(cds, chan)
	struct cbus_dma_state *cds;
	int chan;
{

	return (0);
}

int
_cbus_dmamap_create(struct cbus_dma_state *cds, int chan, bus_size_t size,
    int flags)
{
	bus_size_t maxsize;
	int error;

	if (chan < 0 || chan >= DMA_NCHAN) {
		printf("%s: bogus drq %d\n", cds->cds_dev->dv_xname, chan);
		return (EINVAL);
	}

	if (chan & 4)
		maxsize = (1 << 17);
	else
		maxsize = (1 << 16);

	if (size > maxsize)
		return (EINVAL);

	if (CBUS_DMA_DRQ_ISFREE(cds, chan) == 0) {
		printf("%s: drq %d is not free\n", cds->cds_dev->dv_xname,
		    chan);
		return (EAGAIN);
	}

	CBUS_DMA_DRQ_ALLOC(cds, chan);

	error = bus_dmamap_create(cds->cds_dmat, size, 1, size, maxsize,
	    flags, &cds->cds_dmamaps[chan]);

	if (error)
		CBUS_DMA_DRQ_FREE(cds, chan);

	return (error);
}

void
_cbus_dmamap_destroy(struct cbus_dma_state *cds, int chan)
{

	if (chan < 0 || chan >= DMA_NCHAN) {
		printf("%s: bogus drq %d\n", cds->cds_dev->dv_xname, chan);
		goto lose;
	}

	if (CBUS_DMA_DRQ_ISFREE(cds, chan)) {
		printf("%s: drq %d is already free\n",
		    cds->cds_dev->dv_xname, chan);
		goto lose;
	}

	CBUS_DMA_DRQ_FREE(cds, chan);

	bus_dmamap_destroy(cds->cds_dmat, cds->cds_dmamaps[chan]);

	if ((chan & 0x04) == 0)
		bus_space_write_1(cds->cds_bst, cds->cds_dmaabah, 0, 
				  DMAABA_16M | chan);

	return;

 lose:
	panic("_cbus_dmamap_destroy");
}

/*
 * _cbus_dmastart(): program 8237 DMA controller channel and set it
 * in motion.
 */
int
_cbus_dmastart(struct cbus_dma_state *cds, int chan,
    void *addr, bus_size_t nbytes, struct proc *p, int flags, int busdmaflags)
{
	bus_dmamap_t dmam;
	bus_addr_t dmaaddr;
	int waport;
	int ochan = chan & 3;
	int error;

	if (chan < 0 || chan >= DMA_NCHAN) {
		printf("%s: bogus drq %d\n", cds->cds_dev->dv_xname, chan);
		goto lose;
	}

#ifdef CBUSDMA_DEBUG
	printf("_cbus_dmastart: drq %d, addr %p, nbytes 0x%lx, p %p, "
	    "flags 0x%x, dmaflags 0x%x\n",
	    chan, addr, nbytes, p, flags, busdmaflags);
#endif

	if (chan & 4) {
		if (nbytes > (1 << 17) || nbytes & 1 || (u_long)addr & 1) {
			printf("%s: drq %d, nbytes 0x%lx, addr %p\n",
			    cds->cds_dev->dv_xname, chan, nbytes, addr);
			goto lose;
		}
	} else {
		if (nbytes > (1 << 16)) {
			printf("%s: drq %d, nbytes 0x%lx\n",
			    cds->cds_dev->dv_xname, chan, nbytes);
			goto lose;
		}
	}

	dmam = cds->cds_dmamaps[chan];
	if (dmam == NULL)
		panic("_cbus_dmastart: no DMA map for chan %d\n", chan);

	error = bus_dmamap_load(cds->cds_dmat, dmam, addr, nbytes,
	    p, busdmaflags);
	if (error)
		return (error);

#ifdef CBUSDMA_DEBUG
	__asm(".globl cbus_dmastart_afterload ; cbus_dmastart_afterload:");
#endif

	if (flags & DMAMODE_READ) {
		bus_dmamap_sync(cds->cds_dmat, dmam, 0, dmam->dm_mapsize,
		    BUS_DMASYNC_PREREAD);
		cds->cds_dmareads |= (1 << chan);
	} else {
		bus_dmamap_sync(cds->cds_dmat, dmam, 0, dmam->dm_mapsize,
		    BUS_DMASYNC_PREWRITE);
		cds->cds_dmareads &= ~(1 << chan);
	}

	dmaaddr = dmam->dm_segs[0].ds_addr;

#ifdef CBUSDMA_DEBUG
	printf("     dmaaddr 0x%lx\n", dmaaddr);

	__asm(".globl cbus_dmastart_aftersync ; cbus_dmastart_aftersync:");
#endif

	cds->cds_dmalength[chan] = nbytes;

	_cbus_dmamask(cds, chan);
	cds->cds_dmafinished &= ~(1 << chan);

	if ((chan & 4) == 0) {
		if (dmamode[flags] & DMA37MD_LOOP)
			bus_space_write_1(cds->cds_bst, cds->cds_dmaabah, 0, 
					  DMAABA_64K | chan); 
		else
			bus_space_write_1(cds->cds_bst, cds->cds_dmaabah, 0,
					  DMAABA_16M | chan);

		/* set dma channel mode */
		bus_space_write_1(cds->cds_bst, cds->cds_dma1h, DMA1_MODE,
		    ochan | dmamode[flags]);

		/* send start address */
		waport = DMA1_CHN(ochan);
		bus_space_write_1(cds->cds_bst, cds->cds_dmapgh,
		    dmapageport[0][ochan], (dmaaddr >> 16) & 0xff);
		bus_space_write_1(cds->cds_bst, cds->cds_dma1h, waport,
		    dmaaddr & 0xff);
		bus_space_write_1(cds->cds_bst, cds->cds_dma1h, waport,
		    (dmaaddr >> 8) & 0xff);

		/* send count */
		bus_space_write_1(cds->cds_bst, cds->cds_dma1h, waport + 1,
		    (--nbytes) & 0xff);
		bus_space_write_1(cds->cds_bst, cds->cds_dma1h, waport + 1,
		    (nbytes >> 8) & 0xff);
	} else {
		/* set dma channel mode */
		bus_space_write_1(cds->cds_bst, cds->cds_dma2h, DMA2_MODE,
		    ochan | dmamode[flags]);

		/* send start address */
		waport = DMA2_CHN(ochan);
		bus_space_write_1(cds->cds_bst, cds->cds_dmapgh,
		    dmapageport[1][ochan], (dmaaddr >> 16) & 0xff);
		dmaaddr >>= 1;
		bus_space_write_1(cds->cds_bst, cds->cds_dma2h, waport,
		    dmaaddr & 0xff);
		bus_space_write_1(cds->cds_bst, cds->cds_dma2h, waport,
		    (dmaaddr >> 8) & 0xff);

		/* send count */
		nbytes >>= 1;
		bus_space_write_1(cds->cds_bst, cds->cds_dma2h, waport + 2,
		    (--nbytes) & 0xff);
		bus_space_write_1(cds->cds_bst, cds->cds_dma2h, waport + 2,
		    (nbytes >> 8) & 0xff);
	}

	_cbus_dmaunmask(cds, chan);
	return (0);

 lose:
	panic("_cbus_dmastart");
}

void
_cbus_dmaabort(struct cbus_dma_state *cds, int chan)
{

	if (chan < 0 || chan >= DMA_NCHAN) {
		printf("%s: bogus drq %d\n", cds->cds_dev->dv_xname, chan);
		panic("_cbus_dmaabort");
	}

	_cbus_dmamask(cds, chan);
	bus_dmamap_unload(cds->cds_dmat, cds->cds_dmamaps[chan]);
	cds->cds_dmareads &= ~(1 << chan);
}

bus_size_t
_cbus_dmacount(struct cbus_dma_state *cds, int chan)
{
	int waport;
	bus_size_t nbytes;
	int ochan = chan & 3;

	if (chan < 0 || chan >= DMA_NCHAN) {
		printf("%s: bogus drq %d\n", cds->cds_dev->dv_xname, chan);
		panic("cbus_dmacount");
	}

	_cbus_dmamask(cds, chan);

	/*
	 * We have to shift the byte count by 1.  If we're in auto-initialize
	 * mode, the count may have wrapped around to the initial value.  We
	 * can't use the TC bit to check for this case, so instead we compare
	 * against the original byte count.
	 * If we're not in auto-initialize mode, then the count will wrap to
	 * -1, so we also handle that case.
	 */
	if ((chan & 4) == 0) {
		waport = DMA1_CHN(ochan);
		nbytes = bus_space_read_1(cds->cds_bst, cds->cds_dma1h,
		    waport + 1) + 1;
		nbytes += bus_space_read_1(cds->cds_bst, cds->cds_dma1h,
		    waport + 1) << 8;
		nbytes &= 0xffff;
	} else {
		waport = DMA2_CHN(ochan);
		nbytes = bus_space_read_1(cds->cds_bst, cds->cds_dma2h,
		    waport + 2) + 1;
		nbytes += bus_space_read_1(cds->cds_bst, cds->cds_dma2h,
		    waport + 2) << 8;
		nbytes <<= 1;
		nbytes &= 0x1ffff;
	}

	if (nbytes == cds->cds_dmalength[chan])
		nbytes = 0;

	_cbus_dmaunmask(cds, chan);
	return (nbytes);
}

int
_cbus_dmafinished(struct cbus_dma_state *cds, int chan)
{

	if (chan < 0 || chan >= DMA_NCHAN) {
		printf("%s: bogus drq %d\n", cds->cds_dev->dv_xname, chan);
		panic("_cbus_dmafinished");
	}

	/* check that the terminal count was reached */
	if ((chan & 4) == 0)
		cds->cds_dmafinished |= bus_space_read_1(cds->cds_bst,
		    cds->cds_dma1h, DMA1_SR) & 0x0f;
	else
		cds->cds_dmafinished |= (bus_space_read_1(cds->cds_bst,
		    cds->cds_dma2h, DMA2_SR) & 0x0f) << 4;

	return ((cds->cds_dmafinished & (1 << chan)) != 0);
}

void
_cbus_dmadone(struct cbus_dma_state *cds, int chan)
{
	bus_dmamap_t dmam;

	if (chan < 0 || chan >= DMA_NCHAN) {
		printf("%s: bogus drq %d\n", cds->cds_dev->dv_xname, chan);
		panic("_cbus_dmadone");
	}

	dmam = cds->cds_dmamaps[chan];

	_cbus_dmamask(cds, chan);

	if (_cbus_dmafinished(cds, chan) == 0)
		printf("%s: _cbus_dmadone: channel %d not finished\n",
		    cds->cds_dev->dv_xname, chan);

	bus_dmamap_sync(cds->cds_dmat, dmam, 0, dmam->dm_mapsize,
	    (cds->cds_dmareads & (1 << chan)) ? BUS_DMASYNC_POSTREAD :
	    BUS_DMASYNC_POSTWRITE);

	bus_dmamap_unload(cds->cds_dmat, dmam);
	cds->cds_dmareads &= ~(1 << chan);
}

void
_cbus_dmafreeze(struct cbus_dma_state *cds)
{

	panic("cbusdma: freeze not implemented");
}

void
_cbus_dmathaw(struct cbus_dma_state *cds)
{

	panic("cbusdma: freeze not implemented");
}

int
_cbus_dmamem_alloc(struct cbus_dma_state *cds, int chan,
    bus_size_t size, bus_addr_t *addrp, int flags)
{
	bus_dma_segment_t seg;
	int error, boundary, rsegs;

	if (chan < 0 || chan >= DMA_NCHAN) {
		printf("%s: bogus drq %d\n", cds->cds_dev->dv_xname, chan);
		panic("_cbus_dmamem_alloc");
	}

	boundary = (chan & 4) ? (1 << 17) : (1 << 16);

	size = round_page(size);

	error = bus_dmamem_alloc(cds->cds_dmat, size, NBPG, boundary,
	    &seg, 1, &rsegs, flags);
	if (error)
		return (error);

	*addrp = seg.ds_addr;
	return (0);
}

void
_cbus_dmamem_free(struct cbus_dma_state *cds, int chan,
    bus_addr_t addr, bus_size_t size)
{
	bus_dma_segment_t seg;

	if (chan < 0 || chan >= DMA_NCHAN) {
		printf("%s: bogus drq %d\n", cds->cds_dev->dv_xname, chan);
		panic("_cbus_dmamem_free");
	}

	seg.ds_addr = addr;
	seg.ds_len = size;

	bus_dmamem_free(cds->cds_dmat, &seg, 1);
}

int
_cbus_dmamem_map(struct cbus_dma_state *cds, int chan,
    bus_addr_t addr, bus_size_t size, caddr_t *kvap)
	int flags;
{
	bus_dma_segment_t seg;

	if (chan < 0 || chan >= DMA_NCHAN) {
		printf("%s: bogus drq %d\n", cds->cds_dev->dv_xname, chan);
		panic("_cbus_dmamem_map");
	}

	seg.ds_addr = addr;
	seg.ds_len = size;

	return (bus_dmamem_map(cds->cds_dmat, &seg, 1, size, kvap, flags));
}

void
_cbus_dmamem_unmap(struct cbus_dma_state *cds, int chan,
    caddr_t kva, size_t size)
{

	if (chan < 0 || chan >= DMA_NCHAN) {
		printf("%s: bogus drq %d\n", cds->cds_dev->dv_xname, chan);
		panic("_cbus_dmamem_unmap");
	}

	bus_dmamem_unmap(cds->cds_dmat, kva, size);
}

int
_cbus_dmamem_mmap(struct cbus_dma_state *cds, int chan,
    bus_addr_t addr, bus_size_t size, int off, int prot, int flags)
{
	bus_dma_segment_t seg;

	if (chan < 0 || chan >= DMA_NCHAN) {
		printf("%s: bogus drq %d\n", cds->cds_dev->dv_xname, chan);
		panic("_cbus_dmamem_mmap");
	}

	if (off < 0)
		return (-1);

	seg.ds_addr = addr;
	seg.ds_len = size;

	return (bus_dmamem_mmap(cds->cds_dmat, &seg, 1, off, prot, flags));
}

int
_cbus_drq_isfree(struct cbus_dma_state *cds, int chan)
{

	if (chan < 0 || chan >= DMA_NCHAN) {
		printf("%s: bogus drq %d\n", cds->cds_dev->dv_xname, chan);
		panic("_cbus_drq_isfree");
	}

	return CBUS_DMA_DRQ_ISFREE(cds, chan);
}

void *
_cbus_malloc(struct cbus_dma_state *cds, int chan,
    size_t size, int pool, int flags)
{
	bus_addr_t addr;
	caddr_t kva;
	int bflags;
	struct cbus_mem *m;

	bflags = flags & M_WAITOK ? BUS_DMA_WAITOK : BUS_DMA_NOWAIT;

	if (_cbus_dmamem_alloc(cds, chan, size, &addr, bflags))
		return 0;
	if (_cbus_dmamem_map(cds, chan, addr, size, &kva, bflags)) {
		_cbus_dmamem_free(cds, chan, addr, size);
		return 0;
	}
	m = malloc(sizeof(*m), pool, flags);
	if (m == 0) {
		_cbus_dmamem_unmap(cds, chan, kva, size);
		_cbus_dmamem_free(cds, chan, addr, size);
		return 0;
	}
	m->cds = cds;
	m->chan = chan;
	m->size = size;
	m->addr = addr;
	m->kva = kva;
	m->next = cbus_mem_head;
	cbus_mem_head = m;
	return (void *)kva;
}

void
_cbus_free(void *addr, int pool)
{
	struct cbus_mem **mp, *m;
	caddr_t kva = (caddr_t)addr;

	for(mp = &cbus_mem_head; *mp && (*mp)->kva != kva;
	    mp = &(*mp)->next)
		continue;
	m = *mp;
	if (m == NULL) {
		printf("_cbus_free: freeing unallocted memory\n");
		return;
	}
	*mp = m->next;
	_cbus_dmamem_unmap(m->cds, m->chan, kva, m->size);
	_cbus_dmamem_free(m->cds, m->chan, m->addr, m->size);
	free(m, pool);
}

int
_cbus_mappage(mem, off, prot)
	void *mem;
	int off;
	int prot;
{
	struct cbus_mem *m;

	for (m = cbus_mem_head; m && m->kva != (caddr_t)mem; m = m->next)
		coutinue;
	if (m == NULL) {
		printf("_cbus_mappage: mapping unallocted memory\n");
		return -1;
	}
	return _cbus_dmamem_mmap(m->cds, m->chan, m->addr,
	    m->size, off, prot, BUS_DMA_WAITOK);
}
